# Copyright 2021 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import re

from neutron_lib import constants as lib_constants
from neutron_tempest_plugin.common import ssh
from neutron_tempest_plugin import config
from neutron_tempest_plugin.scenario import base
from neutron_tempest_plugin.scenario import constants as neutron_constants
from oslo_log import log
from tempest.common import waiters
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators

from whitebox_neutron_tempest_plugin.common import constants
from whitebox_neutron_tempest_plugin.common import utils
from whitebox_neutron_tempest_plugin.tests.scenario import base as ds_base


CONF = config.CONF
LOG = log.getLogger(__name__)
IPV4 = lib_constants.IP_VERSION_4
IPV6 = lib_constants.IP_VERSION_6


class ExtraDhcpOptionsTest(base.BaseTempestTestCase):
    credentials = ['primary', 'admin']
    required_extensions = ['extra_dhcp_opt']
    ipv4_cidr_pattern = '192.168.{}.0/24'
    ipv6_cidr_pattern = '2001:{:x}::/64'

    @classmethod
    def resource_setup(cls):
        super(ExtraDhcpOptionsTest, cls).resource_setup()
        cls.rand_name = data_utils.rand_name(
            cls.__name__.rsplit('.', 1)[-1])
        cls.keypair = cls.create_keypair(name=cls.rand_name)
        cls.security_group = cls.create_security_group(name=cls.rand_name)
        cls.create_loginable_secgroup_rule(
            cls.security_group['id'], client=cls.client)

        if CONF.neutron_plugin_options.default_image_is_advanced:
            cls.flavor_ref = CONF.compute.flavor_ref
            cls.image_ref = CONF.compute.image_ref
            cls.username = CONF.validation.image_ssh_user
        else:
            cls.flavor_ref = \
                CONF.neutron_plugin_options.advanced_image_flavor_ref
            cls.image_ref = CONF.neutron_plugin_options.advanced_image_ref
            cls.username = CONF.neutron_plugin_options.advanced_image_ssh_user

        if (not cls.flavor_ref) or (not cls.image_ref):
            raise cls.skipException(
                'No advance image/flavor available for these tests')

    def setUp(self):
        super(ExtraDhcpOptionsTest, self).setUp()
        self.subnet = None
        self.port = None
        self.server = None
        self.dhcp_disabled = {}

    def _check_created_extra_dhcp_opts(self):
        # obtain the extra_dhcp_opts from Neutron API
        retrieved = self.client.show_port(
            self.port['id'])['port']['extra_dhcp_opts']
        self.assertEqual(len(retrieved), len(self.extra_dhcp_opts))
        for retrieved_option in retrieved:
            for option in self.extra_dhcp_opts:
                # default ip_version is 4
                ip_version = option.get('ip_version', IPV4)
                if (retrieved_option['opt_value'] == option['opt_value'] and
                        retrieved_option['opt_name'] == option['opt_name'] and
                        retrieved_option['ip_version'] == ip_version):
                    break
            else:
                self.fail('Extra DHCP option not found in port %s' %
                          str(retrieved_option))

    def _create_port_and_check_dhcp_opts(
            self,
            ipv6=False,
            ra_address_mode=None,
            dhcp4_enabled=True,
            dhcp6_enabled=True):
        rand_name_test = self.rand_name + '-' + self._testMethodName
        network = self.create_network(name=rand_name_test)

        subnet_index = len(self.reserved_subnet_cidrs)
        self.subnet = self.create_subnet(
            network=network,
            cidr=self.ipv4_cidr_pattern.format(subnet_index),
            name=rand_name_test,
            enable_dhcp=dhcp4_enabled)
        router = self.create_router_by_client()
        self.create_router_interface(router['id'], self.subnet['id'])
        if ipv6:
            if dhcp6_enabled:
                subnetv6 = self.create_subnet(
                    network=network,
                    ip_version=IPV6,
                    name=rand_name_test + '-ipv6',
                    cidr=self.ipv6_cidr_pattern.format(subnet_index),
                    ipv6_ra_mode=ra_address_mode,
                    ipv6_address_mode=ra_address_mode)
            else:
                # when dhcp6 is disabled, neither ipv6_ra_mode nor
                # ipv6_address_mode should be provided
                subnetv6 = self.create_subnet(
                    network=network,
                    ip_version=IPV6,
                    name=rand_name_test + '-ipv6',
                    cidr=self.ipv6_cidr_pattern.format(subnet_index),
                    enable_dhcp=dhcp6_enabled)
            self.create_router_interface(router['id'], subnetv6['id'])

        # a port is created with the extra_dhcp_options from the test case
        self.port = self.create_port(
            network=network, name=rand_name_test,
            security_groups=[self.security_group['id']],
            extra_dhcp_opts=self.extra_dhcp_opts)

        self._check_created_extra_dhcp_opts()

    def _create_server_and_fip(self):
        networks = [{'port': self.port['id']}]
        # config_drive is needed when dhcp4 is disabled in order to provide
        # an IPv4 to the server to be able to connect via ssh to it
        config_drive = not self.subnet['enable_dhcp']
        self.server = self.create_server(
            flavor_ref=self.flavor_ref,
            image_ref=self.image_ref,
            key_name=self.keypair['name'],
            networks=networks,
            config_drive=config_drive,
            name=self.rand_name + '-' + self._testMethodName)['server']

        vm_fip = self.create_floatingip(port=self.port)['floating_ip_address']
        return ssh.Client(vm_fip,
                          self.username,
                          pkey=self.keypair['private_key'])

    def _test_extra_dhcp_opts_ipv4_ipv6(self, ra_address_mode):
        self._create_port_and_check_dhcp_opts(
            ipv6=True, ra_address_mode=ra_address_mode)
        vm_ssh_client = self._create_server_and_fip()

        nmcli_dhcp_options = {}
        nmcli_dhcp_options[IPV4] = utils.parse_dhcp_options_from_nmcli(
            vm_ssh_client, IPV4)
        nmcli_dhcp_options[IPV6] = utils.parse_dhcp_options_from_nmcli(
            vm_ssh_client, IPV6, timeout=100)

        for extra_dhcp_opt in self.extra_dhcp_opts:
            ip_version = extra_dhcp_opt.get('ip_version', IPV4)

            # if opt_name is an integer value, then we need to obtain its
            # corresponding option name
            try:
                int(extra_dhcp_opt['opt_name'])
            except ValueError:
                # it is not an integer
                opt_name = extra_dhcp_opt['opt_name']
            else:
                opt_name = constants.DHCP_OPTIONS_NUMBER_TO_NAME[
                    extra_dhcp_opt['opt_name']]

            self.assertIn(opt_name, nmcli_dhcp_options[ip_version])
            # some formatting is needed before the value comparison:
            # - remove extra " from Neutron option
            # - remove extra \ from nmcli option
            self.assertEqual(
                extra_dhcp_opt['opt_value'].replace('"', ''),
                nmcli_dhcp_options[ip_version][opt_name].replace('\\', ''))

    @decorators.idempotent_id('8f52b4dc-faae-4f1d-b113-d2f3e86bf0d6')
    def test_extra_dhcp_opts_ipv4_ipv6_stateful(self):
        self.extra_dhcp_opts = [
            {'opt_value': '2001:4860:4860::8888',
             'opt_name': 'dns-server',
             'ip_version': IPV6},
            {'opt_value': '8.8.8.8',
             'opt_name': 'dns-server',
             'ip_version': IPV4},
            {'opt_value': '1600',
             'opt_name': '26'}]  # 26 is option code for mtu

        self._test_extra_dhcp_opts_ipv4_ipv6('dhcpv6-stateful')

    @decorators.idempotent_id('e9e32249-6148-4565-b7b1-e64c77c9f4ec')
    def test_extra_dhcp_opts_ipv4_ipv6_stateless(self):
        self.extra_dhcp_opts = [
            {'opt_value': '"ipv6.domain"',  # domains should be between "
             'opt_name': 'domain-search',
             'ip_version': IPV6},
            {'opt_value': '"ipv4.domain"',  # ditto
             'opt_name': 'domain-name',
             'ip_version': IPV4}]

        self._test_extra_dhcp_opts_ipv4_ipv6('dhcpv6-stateless')

    @decorators.idempotent_id('ef41d6d8-f2bf-44e4-9f4d-bb8a3fed50ad')
    def test_extra_dhcp_opts_disabled_enabled_dhcp4(self):
        domain_value = 'ipv4.domain'
        domain_opt = 'domain-name'
        self.extra_dhcp_opts = [
            {'opt_value': '"{}"'.format(domain_value),
             'opt_name': domain_opt,
             'ip_version': IPV4}]

        # dhcp is disabled from the iPv4 subnet created
        self._create_port_and_check_dhcp_opts(dhcp4_enabled=False)
        vm_ssh_client = self._create_server_and_fip()
        # ipv4.domain is not expected
        vm_resolv_conf = vm_ssh_client.exec_command(
            "sudo dhclient && cat /etc/resolv.conf")
        self.assertIsNone(re.search(r'^search\s+{}\s+'.format(domain_value),
                                    vm_resolv_conf,
                                    re.MULTILINE))

        # enable dhcp for the IPv4 subnet and reboot the VM
        self.subnets[-1] = self.client.update_subnet(
            self.subnets[-1]['id'], enable_dhcp=True)['subnet']
        self.os_primary.servers_client.reboot_server(
            self.server['id'], type='SOFT')
        waiters.wait_for_server_status(self.os_primary.servers_client,
                                       self.server['id'],
                                       neutron_constants.SERVER_STATUS_ACTIVE)
        # ipv4.domain is expected
        vm_resolv_conf = vm_ssh_client.exec_command(
            "sudo dhclient && cat /etc/resolv.conf")
        # (rsafrono) this regex will work reliably even in case dhcp_domain in
        # nova.conf is defined, i.e. not an empty value. Helps to stabilize
        # the test on podified environments.
        self.assertIsNotNone(re.search(r'^search\s+{}\s+'.format(domain_value),
                                       vm_resolv_conf,
                                       re.MULTILINE))

    @decorators.idempotent_id('abb12899-690a-407d-99d4-49eca030ce94')
    def test_extra_dhcp_opts_disabled_dhcp6(self):
        domain_value = 'ipv6.domain'
        domain_opt = 'domain-search'
        self.extra_dhcp_opts = [
            {'opt_value': '"{}"'.format(domain_value),
             'opt_name': domain_opt,
             'ip_version': IPV6}]

        # dhcp is disabled from the iPv6 subnet created
        self._create_port_and_check_dhcp_opts(ipv6=True, dhcp6_enabled=False)
        vm_ssh_client = self._create_server_and_fip()

        # empty nmcli_dhcp6_options is expected
        nmcli_dhcp6_options = utils.parse_dhcp_options_from_nmcli(
            vm_ssh_client, IPV6, timeout=100, expected_empty=True)
        self.assertIsNone(nmcli_dhcp6_options)


class OvnExtraDhcpOptionsTest(ExtraDhcpOptionsTest,
                              ds_base.BaseTempestTestCaseOvn):
    def _check_created_extra_dhcp_opts(self):
        def _check_options_from_ovn_nbdb(dhcp_option_uuids):
            cmd2_pattern = '{} get dhcp_options {} options:{}'

            # are dhcpv4 and dhcpv6 enabled?
            subnets_dhcp_enabled = {subnet['ip_version']: subnet['enable_dhcp']
                                    for subnet in self.subnets}

            for extra_dhcp_opt in self.extra_dhcp_opts:
                ip_version = extra_dhcp_opt.get('ip_version', IPV4)
                if (not subnets_dhcp_enabled[ip_version] or
                        self.dhcp_disabled.get(ip_version) is True or
                        extra_dhcp_opt['opt_name'] == 'dhcp_disabled'):
                    continue
                ovn_opt_name = constants.DHCP_OPTIONS_NEUTRON_TO_OVN[
                    extra_dhcp_opt['opt_name']]
                # if the option is not found in the OVN DB, the command will
                # fail and then the test will fail, which is correct
                ovn_opt_value = self.run_on_master_controller(
                    cmd2_pattern.format(
                        self.nbctl,
                        dhcp_option_uuids[ip_version],
                        ovn_opt_name))
                # some formatting is needed before the value comparison:
                # - remove extra " from Neutron option
                # - remove extra \ and " from OVN option
                self.assertEqual(
                    extra_dhcp_opt['opt_value'].replace('"', ''),
                    ovn_opt_value.replace('"', '').replace('\\', '').strip())

            # Verify no dhcpvX_option_uuid exist when DHCPvX is disabled
            # Value '[]' means no entry in OVN NBDB dhcp_options table
            for ip_version in subnets_dhcp_enabled:
                if (subnets_dhcp_enabled[ip_version] is False or
                        self.dhcp_disabled.get(ip_version) is True):
                    self.assertEqual('[]', dhcp_option_uuids[ip_version])

        # first, call parent's method
        super(OvnExtraDhcpOptionsTest, self)._check_created_extra_dhcp_opts()

        cmd1_pattern = '{} find logical_switch_port name={}'
        output = self.run_on_master_controller(
            cmd1_pattern.format(self.nbctl, self.port['id']))
        dhcp_option_uuids = {}

        for line in output.splitlines():
            if re.search(r'^dhcpv4_options', line):
                ip_version = IPV4
            elif re.search(r'^dhcpv6_options', line):
                ip_version = IPV6
            else:
                # skip this line
                continue
            dhcp_option_uuids[ip_version] = re.sub(
                r'^dhcpv{}_options\s+: '.format(ip_version),
                '',
                line).strip()

        _check_options_from_ovn_nbdb(dhcp_option_uuids)

    @decorators.idempotent_id('30ef3221-e46b-4358-b550-d98c08272e1c')
    def test_extra_dhcp_opts_dhcp_disabled_option(self):
        self.extra_dhcp_opts = [
            {'opt_value': '2001:4860:4860::8888',
             'opt_name': 'dns-server',
             'ip_version': IPV6},
            {'opt_value': '8.8.8.8',
             'opt_name': 'dns-server',
             'ip_version': IPV4},
            {'opt_value': 'True',
             'opt_name': 'dhcp_disabled',
             'ip_version': IPV6},
            {'opt_value': 'False',
             'opt_name': 'dhcp_disabled',
             'ip_version': IPV4}]
        self.dhcp_disabled[IPV6] = True
        self.dhcp_disabled[IPV4] = False

        # create port and check extra-dhcp-opts
        self._create_port_and_check_dhcp_opts(
            ipv6=True, ra_address_mode='dhcpv6-stateless')

        # update port and check extra-dhcp-opts
        self.extra_dhcp_opts = [
            {'opt_value': '2001:4860:4860::8888',
             'opt_name': 'dns-server',
             'ip_version': IPV6},
            {'opt_value': '8.8.8.8',
             'opt_name': 'dns-server',
             'ip_version': IPV4},
            {'opt_value': 'False',
             'opt_name': 'dhcp_disabled',
             'ip_version': IPV6},
            {'opt_value': 'True',
             'opt_name': 'dhcp_disabled',
             'ip_version': IPV4}]
        self.dhcp_disabled[IPV6] = False
        self.dhcp_disabled[IPV4] = True
        self.port = self.client.update_port(
            self.port['id'], extra_dhcp_opts=self.extra_dhcp_opts)['port']
        self._check_created_extra_dhcp_opts()
